/*
 * clock.c - 60LED clock for me
 *   by uratan! 2014.11.1
 */
#include "common.h"
#include <chip.h>

/* =============================================================== */

/*
 * version of this firmware
 *   (forked from wsmon-020)
 *      (21 <== 2.1 <== 021)
 */
#define VERSION   29

/*
 * state for clock operations
 */
#define S_DISPLAY    0		/* display the time */
#define S_TUNE_ORIG  1		/* change position of origin */
#define S_TUNE_HOUR  2		/* change hour */
#define S_TUNE_MIN   3		/* change minute */

/*
 * definition for incoming event
 */
#define PUSH_SHORT    'S'
#define PUSH_LONG     'L'

/* =============================================================== */

/*
 * typedefs and variables, and related macros
 */
/*
 * indicate position of "twelve o'clock"
 */
int clk_origin;			/* has 0--59 */

/*
 * for RTC operation
 */
typedef struct {
	u_long hour;			/* has 0--23 */
	u_long minute;			/*     0--59 */
	u_long second;			/*     0--59 */
	u_long under_sec;		/*     0--59 */
} T_TIME;

T_TIME clk_curr_time;		/* has the time at now */

/*
 * typedef for brightness values
 *   |<--------- long ---------->|
 *   +------+------+------+------+
 *   | dumm | green| red  | blue |
 *   +------+------+------+------+
 */
typedef union {
	u_long grb;
	struct {
		u_char b;		/* because his endian is little */
		u_char r;
		u_char g;
		u_char dumm;
	} c;
	struct {
		u_short A;
		u_short B;
	} brt;
} T_CODE_GRB_brt;

#define N_LEDs   60
T_CODE_GRB_brt clk_code_grb[N_LEDs];

/*
 * macro to take out correct index avoiding division
 */
//#define clip_index(var)    (var % 60)
#define clamp_index(var)    do{ while(var >= 60) { var -= 60; } }while(0)

/*
 * to handle color variation
 */
typedef struct {
	u_long hand;
	u_long dial;
} T_COLOR_PAIR;

#define N_COLORS      24
#define COLORS_BASE  0x04		/* handle 0x04 as 1 */
#define COLORS_MASK  0x3f3f3f	/* mask fraction bits after division */
#define MAX_BRIGHTNESS  (256 / COLORS_BASE - 1)
		/*
		 * you shall use brightness values (for dials) in this range:
		 *     min:  4 ... lesser than this is rounded to 0
		 *     max: 63 ... greater then this will make overflow
		 */
int clk_disp_color;
const T_COLOR_PAIR clk_color_pair[N_COLORS] = {
			/*  hands     dials */
			/*  G/R/B     G/R/B */
/*  0 */	{ 0x040404, 0x000004 },		/* white / blue */
/*  1 */	{ 0x040404, 0x000104 },		/*       / ... */
/*  2 */	{ 0x040404, 0x000204 },		/*       / ... */
/*  3 */	{ 0x040404, 0x000304 },		/*       / ... */
/*  4 */	{ 0x040404, 0x000404 },		/*       / purple */
/*  5 */	{ 0x040404, 0x000403 },		/*       / ... */
/*  6 */	{ 0x040404, 0x000402 },		/*       / ... */
/*  7 */	{ 0x040404, 0x000401 },		/*       / ... */
/*  8 */	{ 0x040404, 0x000400 },		/*       / red */
/*  9 */	{ 0x040404, 0x010400 },		/*       / ... */
/* 10 */	{ 0x040404, 0x020400 },		/*       / ... */
/* 11 */	{ 0x040404, 0x030400 },		/*       / ... */
/* 12 */	{ 0x040404, 0x040400 },		/*       / yellow */
/* 13 */	{ 0x040404, 0x040300 },		/*       / ... */
/* 14 */	{ 0x040404, 0x040200 },		/*       / ... */
/* 15 */	{ 0x040404, 0x040100 },		/*       / ... */
/* 16 */	{ 0x040404, 0x040000 },		/*       / green */
/* 17 */	{ 0x040404, 0x040001 },		/*       / ... */
/* 18 */	{ 0x040404, 0x040002 },		/*       / ... */
/* 19 */	{ 0x040404, 0x040003 },		/*       / ... */
/* 20 */	{ 0x040404, 0x040004 },		/*       / cyan */
/* 21 */	{ 0x040404, 0x030004 },		/*       / ... */
/* 22 */	{ 0x040404, 0x020004 },		/*       / ... */
/* 23 */	{ 0x040404, 0x010004 },		/*       / ... */
};

/*
 * functions
 */
void clk_main(void);
int clk_control_state(int c_state, int ev);
void clk_count_up(void);
void clk_make_face(int c_state, int m_tick);
void clk_open(int ver_major, int ver_minor);

/*
 * externs
 */
/* in main.c */
extern VOLATILE int systick_raw;
/* in drvie.c */
extern void drv_select_method_SPI(void);
extern void drv_send_code(u_long *m, int N_leds);
/* in uart1.c */
extern VOLATILE int u1_rcv_data;


/* =============================================================== */

/*
 *
 */
void clk_main(void)
{
	int state;
	int do_update;
	int mod_tick, mod_tick_divider;
	int ev;

	/*
	 * confirm drive method
	 */
	drv_select_method_SPI();

	/*
	 * make openning display
	 */
	clk_open(VERSION /10, VERSION %10);

	/*
	 * initial mine
	 */
	clk_origin = 0;
	clk_curr_time.hour = 12;
	clk_curr_time.minute = 0;
	clk_curr_time.second = 0;
	clk_curr_time.under_sec = 0;
	clk_disp_color = 0;

	state = S_DISPLAY;
	do_update = 0;
	mod_tick = 0;
	mod_tick_divider = 0;

	/*
	 * start clock operation
	 */
	while(1) {
		/*
		 * wait for events
		 *   see debug-uart data at first
		 *   because it will come faster (57600bps) than Ir (2400bps)
		 */
		ev = cm_keyin();
		if(ev == -1) {
			/*
			 * see Ir data if uart0 data is absent
			 */
			ev = u1_rcv_data;
			if(ev != -1) {
				/*
				 * something has come
				 */
				u1_rcv_data = -1;
			}
		}

		/*
		 * process the event
		 */
		if(ev != -1) {
			rprintf("ev:%02x\r\n", ev);
			state = clk_control_state(state, ev);
			rprintf("cs:%d\r\n", state);

			/*
			 * always update the LEDs by the event
			 */
			do_update = 1;
		}

		/*
		 * count up the clock (systick = 60Hz)
		 */
		if(systick_raw) {
			systick_raw = 0;

			/*
			 * count up the clock
			 */
			clk_count_up();

			/*
			 * get 15Hz for intensity modurate
			 */
			if(++mod_tick_divider >= 4) {
				mod_tick_divider = 0;

				/*
				 * but it loops from 0 to 15
				 */
				if(++mod_tick >= 16) {
					mod_tick = 0;
				}

				do_update = 1;		/* effective only in misc tuning mode */
			}

			/*
			 * update every 60Hz if on normal clock mode
			 */
			if(state == S_DISPLAY) {
				do_update = 1;			/* may override above 15Hz trigger */
			}
		}

		/*
		 * finally, make clock face
		 */
		if(do_update) {
			do_update = 0;

//DEBUG_HIGH;
			clk_make_face(state, mod_tick);
//DEBUG_LOW;		/* about max 88usec @ 30MHz */
			drv_send_code(&clk_code_grb[0].grb, N_LEDs);
		}
	}

	/* never return */
}

/* =============================================================== */

/*
 * parse the input event, control state and vars
 */
int clk_control_state(int c_state, int ev)
{
	switch(c_state) {
	case S_DISPLAY:						/* display time */
			if(ev == PUSH_LONG) {
				/*
				 * move change mode
				 */
				c_state = S_TUNE_ORIG;
			}
			else if(ev == PUSH_SHORT) {
				/*
				 * change display option, keep this state
				 */
				if(++clk_disp_color >= N_COLORS) {
					clk_disp_color = 0;
				}
			}
			else {
				;		/* ignore, keep this state */
			}
			break;

	case S_TUNE_ORIG:					/* change origin */
			if(ev == PUSH_LONG) {
				/*
				 * change mode
				 */
				c_state = S_TUNE_HOUR;
			}
			else if(ev == PUSH_SHORT) {
				/*
				 * inc clock origin, keep this state
				 */
				if(++clk_origin >= 60) {
					clk_origin = 0;
				}
			}
			else {
				;		/* ignore, keep this state */
			}
			break;

	case S_TUNE_HOUR:					/* change hour */
			if(ev == PUSH_LONG) {
				/*
				 * change mode
				 */
				c_state = S_TUNE_MIN;
			}
			else if(ev == PUSH_SHORT) {
				/*
				 * inc hour, keep this state
				 */
				if(++clk_curr_time.hour >= 24) {
					clk_curr_time.hour = 0;
				}
			}
			else {
				;		/* ignore, keep this state */
			}
			break;

	case S_TUNE_MIN:					/* change minute */
			if(ev == PUSH_LONG) {
				/*
				 * change mode
				 */
				c_state = S_DISPLAY;
			}
			else if(ev == PUSH_SHORT) {
				/*
				 * inc minute, keep this state
				 */
				if(++clk_curr_time.minute >= 60) {
					clk_curr_time.minute = 0;
				}

				/*
				 * reset second/under_sec only by this operation
				 */
				clk_curr_time.second = 0;
				clk_curr_time.under_sec = 0;
			}
			else {
				;		/* ignore, keep this state */
			}
			break;
	}

	return(c_state);
}

/*
 *
 */
void clk_count_up(void)
{
	if(++clk_curr_time.under_sec >= 60) {
		clk_curr_time.under_sec = 0;

		if(++clk_curr_time.second >= 60) {
			clk_curr_time.second = 0;

			if(++clk_curr_time.minute >= 60) {
				clk_curr_time.minute = 0;

				if(++clk_curr_time.hour >= 24) {
					clk_curr_time.hour = 0;
				}
			}
		}
	}
}

/*
 * brightnesses of intensity moduration for minute hand
 */
const u_char clk_modurate[16] = {
	20, 18, 16, 14, 12, 10,  8,  6,  2,  6,  8, 10, 12, 14, 16, 18,
};

/*
 *
 */
void clk_make_face(int c_state, int m_tick)
{
	int i, origin, pos;
	u_long col_hands, col_dials;
	u_char brt;
	int daytime;

	/*
	 * initial brightness, etc.
	 */
#define brightness   brt.A
#define color_sel    brt.B
	for(i=0; i<N_LEDs; i++) {
		clk_code_grb[i].grb = 0x00000000;
	}
	origin = clk_origin;			/* for optimization, I hope */

		/*
		 * take care about drawing priority,
		 * some drawings will be overrided by drawing after
		 */

	/*
	 * draw origin point (only when tuning-origin mode)
	 */
	if(
		((c_state == S_TUNE_ORIG) && (m_tick & 0x04))	/* make blink */
	) {
		pos = origin;
		clamp_index(pos);
		clk_code_grb[pos].brightness = 20;
		clk_code_grb[pos].color_sel = 1;
	}

	/*
	 * draw hour hand (anytime except tuning-origin mode)
	 *   (keep dials color, do not erase dials because this hand is wide)
	 */
	if(
		 (c_state == S_DISPLAY) ||
		 (c_state == S_TUNE_MIN) ||
		((c_state == S_TUNE_HOUR) && (m_tick & 0x04))	/* make blink */
	) {
		pos = origin;
//DEBUG_HIGH;
#if 0
		pos += clk_curr_time.hour * 5 + clk_curr_time.minute / 12;
			/* max 3.36usec @ 30MHz */
#else /* n */
			/* max 1.42usec @ 30MHz */
		/*if(avoid_division)*/ {
			u_long quotient, dividend;

			quotient = 0;
			dividend = clk_curr_time.minute;
			while(dividend >= 12) {
				dividend -= 12;
				quotient++;
			}

			pos += clk_curr_time.hour * 5 + quotient;
		}
#endif /* n */
//DEBUG_LOW;
		pos += 60 - 2;
		clamp_index(pos);
		clk_code_grb[pos++].brightness =  2;		/* -2 */
		clamp_index(pos);
		clk_code_grb[pos++].brightness =  6;		/* -1 */
		clamp_index(pos);
		clk_code_grb[pos++].brightness = 10;		/*  0 */
		clamp_index(pos);
		clk_code_grb[pos++].brightness =  6;		/* +1 */
		clamp_index(pos);
		clk_code_grb[pos++].brightness =  2;		/* +2 */
	}

	/*
	 * draw dials (anytime except tuning-origin mode)
	 */
	if(
		 (c_state == S_DISPLAY) ||
		 (c_state == S_TUNE_MIN) ||
		 (c_state == S_TUNE_HOUR)
	) {
		pos = origin;
		for(i=0; i<12; i++) {
			clamp_index(pos);
			if(i == 0) {
				clk_code_grb[pos].brightness = 10;
			} else {
				clk_code_grb[pos].brightness =  4;
			}
			clk_code_grb[pos].color_sel = 1;

			pos += 5;
		}
	}

	/*
	 * draw second, under_second (only normal clock mode)
	 */
	if(
		 (c_state == S_DISPLAY)
	) {
		pos = origin + clk_curr_time.second;
		clamp_index(pos);
		clk_code_grb[pos].brightness = 2;
		clk_code_grb[pos].color_sel = 0;		/* cancel dials color */

		/*
		 * draw under_second
		 */
		pos = origin + clk_curr_time.under_sec;
		clamp_index(pos);
		clk_code_grb[pos].brightness = 2;
		clk_code_grb[pos].color_sel = 0;		/* cancel dials color */
	}

	/*
	 * draw minute hand (anytime except tuning-origin mode)
	 *   (always intensity modurated)
	 */
	if(
		 (c_state == S_DISPLAY) ||
		 (c_state == S_TUNE_MIN) ||
		 (c_state == S_TUNE_HOUR)
	) {
		pos = origin + clk_curr_time.minute;
		clamp_index(pos);
		clk_code_grb[pos].brightness = clk_modurate[m_tick];
		clk_code_grb[pos].color_sel = 0;		/* cancel dials color */
	}


	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


	/*
	 * select a color and intensity
	 */
	col_hands = clk_color_pair[clk_disp_color].hand;
	col_dials = clk_color_pair[clk_disp_color].dial;
	daytime = ((6 <= clk_curr_time.hour) && (clk_curr_time.hour < 18)) ? 1 : 0;

	/*
	 * convert the brightness to GRB color
	 */
	for(i=0; i<N_LEDs; i++) {
		/*
		 * clip the brightness
		 */
		brt = clk_code_grb[i].brightness;
		if(brt > MAX_BRIGHTNESS) {
			brt = MAX_BRIGHTNESS;
		}

		/*
		 * add colors
		 */
		if(clk_code_grb[i].color_sel) {
			clk_code_grb[i].grb = (col_dials * brt / COLORS_BASE) & COLORS_MASK;
		} else {
			clk_code_grb[i].grb = (col_hands * brt / COLORS_BASE) & COLORS_MASK;
		}

		/*
		 * change intensity by hour
		 */
		if(daytime) {
			clk_code_grb[i].grb *= 2;	/* high intensity while daytime */
		}
	}
}

/* =============================================================== */

/*
 * openning display
 */
void clk_open(int ver_major, int ver_minor)
{
	int ff, t, i;
	int divider;

	divider = 0;

	/*
	 * fade-in all 60 LEDs, then fade-out
	 */
	ff = 0;
	while(ff < (15+1+15)) {
		/*
		 * wait for 60Hz trigger
		 */
		while(! systick_raw) {
			;
		}
		systick_raw = 0;

		/*
		 * make the trigger slow (60Hz / 3 = 20Hz)
		 */
		if(++divider < 3) {
			continue;
		}
		divider = 0;

		/*
		 * drive LEDs
		 */
		t = (ff <= 15) ? ff : (30 - ff);
		for(i=0; i<N_LEDs; i++) {
			clk_code_grb[i].grb = 0x010101 * t;
		}

		drv_send_code(&clk_code_grb[0].grb, N_LEDs);

		ff++;
	}

	/*
	 * make gradation
	 */
	for(i=0; i<N_COLORS; i++) {
		clk_code_grb[18 + i].grb = clk_color_pair[i].dial;
	}

	/*
	 * display version
	 */
	t = 0;
	while(t < 10) {
		/*
		 * wait for 60Hz trigger
		 */
		while(! systick_raw) {
			;
		}
		systick_raw = 0;

		/*
		 * make the trigger slow (60Hz / 12 = 5Hz)
		 */
		if(++divider < 12) {
			continue;
		}
		divider = 0;

		/*
		 * drive LEDs
		 */
		if(t < ver_major) {
			clk_code_grb[1+ t].grb = 0x0f0000;
		}
		if(t < ver_minor) {
			clk_code_grb[59 - t].grb = 0x000f00;
		}
		clk_code_grb[0].grb = 0x00000f;

		drv_send_code(&clk_code_grb[0].grb, N_LEDs);

		t++;
	}

	/*
	 * wait 2sec
	 */
	t = 0;
	while(t < 120) {
		/*
		 * wait for 60Hz trigger
		 */
		while(! systick_raw) {
			;
		}
		systick_raw = 0;

		t++;
	}
}

/* =============================================================== */
